use sdl2::{event::Event, keyboard::Keycode, render::Canvas, video::Window};
use std::{time::Duration, time::Instant};

mod doodad;
mod inputstatus;
mod physics;
mod ship;
mod spriteatlas;

use crate::doodad::Doodad;
use crate::inputstatus::InputStatus;
use crate::physics::PhysVec;
use crate::ship::Ship;
use crate::spriteatlas::SpriteAtlas;

trait Drawable {
	/// Return the world position of this object
	fn position(&self) -> PhysVec;

	/// Get the position of this drawable on the screen
	/// (0, 0) is at top-left corner.
	///
	/// Returned position is this object's center.
	fn screen_position(&self, canvas: &mut Canvas<Window>, c: &Camera) -> PhysVec {
		let win_size = PhysVec::from(canvas.window().size());
		let h = win_size / 2.0;
		return self.position() - c.pos + h;
	}

	// Draw this item on the screen
	fn draw(&self, canvas: &mut Canvas<Window>, sa: &SpriteAtlas, c: &Camera)
		-> Result<(), String>;
}

enum ShipKind {
	Gypsum,
}

impl ShipKind {
	fn sprite(&self) -> &'static str {
		match self {
			Self::Gypsum => "gypsum.png",
		}
	}
}

struct Camera {
	pos: PhysVec,
}

impl Camera {
	fn new() -> Self {
		Camera {
			pos: PhysVec::zero(),
		}
	}
}

static FTL: f64 = 1.0 / 140.0; // frame time limit

struct System {
	decor: Vec<Box<dyn Drawable>>,
}

impl System {
	fn new() -> Self {
		let mut s = System { decor: Vec::new() };

		s.decor.push(Box::new(Doodad::new(PhysVec::zero())));
		s.decor
			.push(Box::new(Doodad::new(PhysVec { x: 100.0, y: 23.0 })));

		return s;
	}

	fn draw(
		&self,
		canvas: &mut Canvas<Window>,
		sa: &SpriteAtlas,
		c: &Camera,
	) -> Result<(), String> {
		for d in &self.decor {
			d.draw(canvas, sa, c)?
		}

		return Ok(());
	}
}

fn main() -> Result<(), String> {
	let sdl_context = sdl2::init()?;
	let video_subsystem = sdl_context.video()?;

	let window = video_subsystem
		.window("SDL2", 640, 480)
		.position_centered()
		.resizable()
		.build()
		.map_err(|e| e.to_string())?;

	let mut canvas = window
		.into_canvas()
		.accelerated()
		.build()
		.map_err(|e| e.to_string())?;
	let texture_creator = canvas.texture_creator();

	let sa = SpriteAtlas::new(&texture_creator)?;
	let mut event_pump = sdl_context.event_pump()?;

	let system = System::new();

	let mut i = InputStatus::new();
	let mut c = Camera::new();

	let mut s = Ship::new(ShipKind::Gypsum);

	//let mut last_frame_time = 0f64;
	let mut frame_start;

	let mut running = true;
	while running {
		frame_start = Instant::now();

		for event in event_pump.poll_iter() {
			match event {
				Event::Quit { .. }
				| Event::KeyDown {
					keycode: Some(Keycode::Escape),
					..
				} => {
					running = false;
				}
				_ => i.update(event),
			}
		}

		c.pos = s.body.pos;

		// Draw
		canvas.clear();
		system.draw(&mut canvas, &sa, &c)?;
		s.draw(&mut canvas, &sa, &c)?;
		canvas.present();

		let frame_time = frame_start.elapsed().as_secs_f64();

		// Wait
		// (limit frame rate)
		if frame_time < FTL {
			std::thread::sleep(Duration::from_secs_f64(FTL - frame_time));
		}

		// Apply input

		let frame_time = frame_start.elapsed().as_secs_f64();
		//last_frame_time = frame_time;

		s.body.tick(frame_time);

		if i.key_thrust {
			s.body.thrust(50.0 * frame_time);
		}

		if i.key_right {
			s.body.rot(-1.0 * frame_time);
		}

		if i.key_left {
			s.body.rot(1.0 * frame_time);
		}
	}

	Ok(())
}
